import logging
import os
import tarfile
import tempfile
import textwrap
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import dp_xgboost as xgb
from dp_xgboost.callback import TrainingCallback
from dp_xgboost.core import Booster, Metric, _deprecate_positional_args
from dp_xgboost.sklearn import XGBModel, _SklObjective

from sarus.pandas import DataFrame
from sarus.sarus import Dataset

Transform = Tuple[str, Dict[str, Any]]
Data = Any
array_like = Any


class XGBClassifier(xgb.XGBClassifier):
    """A class similar to a keras Model and allowing private remote training.

    The sarus.keras.Model class is a wrapper around the
    `tensorflow.keras.Model` class. This class differs from its parent only on
    the `fit` method. The `fit` method accepts a `target_epsilon`.

        - If the specified `target_epsilon` is equal to 0 (default value), then the `Model` class launches a standard keras training on the synthetic data.

        - If the specified `target_epsilon` is strictly greater than 0, then the `Model` class calls the Sarus API to launch a remote private training.
    """

    @_deprecate_positional_args
    def __init__(
        self,
        *,
        objective: _SklObjective = "binary:logistic",
        use_label_encoder: bool = True,
        **kwargs: Any,
    ):
        super().__init__(
            objective=objective, use_label_encoder=use_label_encoder, **kwargs
        )

    def fit(
        self,
        X: Data,
        y: Data,
        target_epsilon: float = 0,
        *,
        sample_weight: Optional[array_like] = None,
        base_margin: Optional[array_like] = None,
        eval_set: Optional[List[Tuple[array_like, array_like]]] = None,
        eval_metric: Optional[Union[str, List[str], Metric]] = None,
        early_stopping_rounds: Optional[int] = None,
        verbose: Optional[bool] = True,
        xgb_model: Optional[Union[Booster, str, XGBModel]] = None,
        sample_weight_eval_set: Optional[List[array_like]] = None,
        base_margin_eval_set: Optional[List[array_like]] = None,
        feature_weights: Optional[array_like] = None,
        callbacks: Optional[List[TrainingCallback]] = None,
    ) -> XGBModel:
        if target_epsilon < 0:
            raise ValueError(
                f"`target_epsilon` must be positive, got {target_epsilon}"
            )

        if target_epsilon == 0:
            return self._fit_local(
                X=X,
                y=y,
                sample_weight=sample_weight,
                base_margin=base_margin,
                eval_set=eval_set,
                eval_metric=eval_metric,
                early_stopping_rounds=early_stopping_rounds,
                verbose=verbose,
                xgb_model=xgb_model,
                sample_weight_eval_set=sample_weight_eval_set,
                base_margin_eval_set=base_margin_eval_set,
                feature_weights=feature_weights,
                callbacks=callbacks,
            )
        else:
            if not isinstance(X, DataFrame):
                raise TypeError(
                    "Expected `x` to be a sarus.pandas.DataFrame when"
                    " `target_epsilon` > 0."
                )
            if eval_set:
                logging.warning(
                    "Ignoring `eval_set` with remote training. "
                    "Not supported yet."
                )

            if sample_weight:
                logging.warning(
                    "Ignoring `sample_weight` with remote training. "
                    "Not supported yet."
                )

            return self._fit_remote(
                X=X,
                y=y,
                target_epsilon=target_epsilon,
                base_margin=base_margin,
                early_stopping_rounds=early_stopping_rounds,
                verbose=verbose,
                xgb_model=xgb_model,
                feature_weights=feature_weights,
                callbacks=callbacks,
            )

    def _fit_local(
        self,
        X: Data,
        y: Data,
        *,
        sample_weight: Optional[array_like] = None,
        base_margin: Optional[array_like] = None,
        eval_set: Optional[List[Tuple[array_like, array_like]]] = None,
        eval_metric: Optional[Union[str, List[str], Metric]] = None,
        early_stopping_rounds: Optional[int] = None,
        verbose: Optional[bool] = True,
        xgb_model: Optional[Union[Booster, str, XGBModel]] = None,
        sample_weight_eval_set: Optional[List[array_like]] = None,
        base_margin_eval_set: Optional[List[array_like]] = None,
        feature_weights: Optional[array_like] = None,
        callbacks: Optional[List[TrainingCallback]] = None,
    ) -> XGBModel:
        if isinstance(X, Dataset):
            X = X._pandas()
            logging.info("Fitting model locally on synthetic data.")

        X = X.astype(float)
        tree = super().fit(
            X,
            y,
            sample_weight=sample_weight,
            base_margin=base_margin,
            early_stopping_rounds=early_stopping_rounds,
            verbose=verbose,
            xgb_model=xgb_model,
            sample_weight_eval_set=sample_weight_eval_set,
            base_margin_eval_set=base_margin_eval_set,
            feature_weights=feature_weights,
            callbacks=callbacks,
        )
        print("Actual privacy consumption (epsilon): 0.0")
        return tree

    def _fit_remote(
        self,
        X: Data,
        y: Data,
        target_epsilon: float,
        *,
        base_margin: Optional[array_like] = None,
        early_stopping_rounds: Optional[int] = None,
        verbose: Optional[bool] = True,
        xgb_model: Optional[Union[Booster, str, XGBModel]] = None,
        feature_weights: Optional[array_like] = None,
        callbacks: Optional[List[TrainingCallback]] = None,
    ) -> XGBModel:
        client = X.dataset.client
        if client is None:
            raise ValueError(
                f"The Sarus Dataset client is None: can not fit "
                f"remotely with `target_epsilon`={target_epsilon}."
            )

        _ = XGBClassifier._make_transform_def(X.dataset.transforms)

        # Remember previous epsilon
        start_eps = X.dataset.epsilon

        # Push the model to the server
        if xgb_model:
            xgb_model_resp = client.session.post(
                f"{client.base_url}/tasks_input_blobs",
                data=xgb_model,
            )
            if xgb_model_resp.status_code > 200:
                raise Exception(
                    "Error while retrieving the model"
                    f"Full Gateway answer was:{xgb_model_resp}"
                )
            xgb_model_id = xgb_model_resp.json()["id"]
        else:
            xgb_model_id = None

        X_columns = X.dataset._filtered_columns()
        y_column = y.dataset._filtered_columns()

        # Launch the fit tasks
        request = client.session.post(
            f"{client.base_url}/fit",
            json={
                "X_columns": X_columns,
                "y_column": y_column,
                "xgb_model_id": xgb_model_id,
                "X_id": X.dataset.id,
                "y_id": y.dataset.id,
                "target_epsilon": target_epsilon,
                "base_margin": base_margin,
                "early_stopping_rounds": early_stopping_rounds,
                "verbose": verbose,
                "feature_weights": feature_weights,
                "objective": self.objective,
                "n_estimators": self.n_estimators,
                "max_depth": self.max_depth,
                "learning_rate": self.learning_rate,
                "booster": self.booster,
                "lambd": self.reg_lambda,
                "base_score": self.base_score,
                "subsample": self.subsample,
                "min_child_weight": self.min_child_weight,
                "nthread": self.n_jobs,
            },
        )
        if request.status_code > 200:
            if request.status_code == 403:
                raise ValueError(
                    "Query failed with the following error: Privacy budget "
                    "limit exceeded"
                )
            raise Exception(
                f"Error while training the model.\
                            Full Gateway answer was:{request}"
            )
        task_id = request.json()["task"]
        if verbose:
            logging.info(f"Fitting task id: {task_id}")

        status = client._poll_training_status(task_id)
        error_message = status.get("error_message", None)
        if error_message:
            raise RuntimeError(
                f"Training failed with the following error:\n"
                f"{textwrap.indent(error_message, '  |')}"
            )

        print(
            f"Actual privacy consumption (epsilon): "
            f"{X.dataset.epsilon-start_eps:.03f}"
        )

        training_status = client._training_status(task_id)
        if "error_message" not in training_status:
            # Set fetched weights to model
            trained_model: XGBModel = client._fetch_xgboost_model(task_id)
            return trained_model

    @staticmethod
    def _make_transform_def(transforms: List[Transform]) -> Callable:
        def transform_def(
            ds: Dataset, features: Optional[Dict] = None
        ) -> Dataset:
            """Build a function to be sent remotely.

            This function should not make use of objects or functions defined
            in the Sarus module to avoid it being listed as a closure by
            cloudpickle.
            """
            func_mapping = {
                "map": lambda ds, params: ds.map(**params),
                "filter": lambda ds, params: ds.filter(**params),
            }
            for name, params in transforms:
                if name in func_mapping:
                    ds = func_mapping[name](ds, params)
                elif name == "split":
                    size = params["end"] - params["start"]
                    ds = ds.skip(params["start"]).take(size)
            return ds

        return transform_def

    @staticmethod
    def _serialize_model(model: xgb.XGBClassifier) -> bytes:
        """Convert a XGBClassifier to compressed archive format."""
        with tempfile.TemporaryDirectory() as _dir:
            model.save_model(_dir)
            with tempfile.TemporaryDirectory() as _second_dir:
                path = os.path.join(_second_dir, "tmpzip")
                with tarfile.open(path, mode="w:gz") as archive:
                    archive.add(_dir, recursive=True, arcname="")
                with open(path, "rb") as f:
                    ret = f.read()
                    return ret



import json
import os
import tempfile
import typing as t
from collections import defaultdict

import pandas as pd
import pyarrow as pa
import sarus_data_spec.protobuf as sp
import sarus_data_spec.status as stt
import sarus_data_spec.storage.typing as storage_typing
import sarus_data_spec.typing as st
from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.constants import (
    ARROW_TASK,
    BIG_DATA_TASK,
    IS_BIG_DATA,
    SCALAR_TASK,
)
from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
from sarus_data_spec.dataspec_validator.typing import DataspecPrivacyPolicy
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.computations.base import BaseComputation
from sarus_data_spec.manager.computations.local.schema import SchemaComputation
from sarus_data_spec.manager.ops.processor.external.external_op import (
    external_implementation,
)
from sarus_data_spec.manager.ops.processor.routing import (
    TransformedDataset,
    TransformedScalar,
)
from sarus_data_spec.manager.ops.source.routing import SourceScalar

import sarus.manager.dataspec_api as api
from sarus.llm.pretrained import TrainState
from sarus.typing import ADMIN_DS, MOCK, PYTHON_TYPE, Client

from .arrow_local import ToArrowComputation
from .arrow_remote import ToArrowComputationOnServer
from .cache_scalar_local import CacheScalarComputation
from .parquet_local import ToParquetComputation
from .value_local import ValueComputation
from .value_remote import ValueComputationOnServer


class SDKManager(Base):
    """The Manager of the SDK running on the client side.

    This Manager has two additional functionalities compared to the
    DelegatingManager manager.

    First, it manages the relationship with the remote server using the API
    endpoints.

    Second, this Manager defines a MOCK version for every DataSpec. The MOCK is
    defined as a smaller version of a DataSpec. In practice, it is a sample of
    SYNTHETIC at the source and MOCKs of transformed DataSpecs are the
    transforms of the MOCKs.

    The MOCK is created and its value computed in the `infer_output_type`
    method. This serves two purposes. First, it provides immediate feedback to
    the user in case of erroneous computation. Second, it allows identifying the
    MOCK's value Python type which is then used by the SDK to instantiate the
    correct DataSpecWrapper type (e.g. instantiate a sarus.pandas.DataFrame if
    the value is a pandas.DataFrame).
    """

    def __init__(
        self,
        storage: storage_typing.Storage,
        protobuf: sp.Manager,
        client: Client,
    ) -> None:
        super().__init__(storage, protobuf)

        self.schema_computation = SchemaComputation(self)
        self.to_parquet_computation = ToParquetComputation(
            self, arrow_remote=ToArrowComputationOnServer(self)
        )
        self.to_arrow_computation = ToArrowComputation(
            self,
            parquet_computation=self.to_parquet_computation,
            arrow_remote=self.to_parquet_computation.remote_arrow,
        )
        self.cache_scalar_computation = CacheScalarComputation(
            self, ValueComputationOnServer(self)
        )
        self.value_computation = ValueComputation(
            self,
            cache_scalar_computation=self.cache_scalar_computation,
            remote_scalar=self.cache_scalar_computation.remote_scalar,
        )

        self._client = client
        self._mock_size = 1000
        self._parquet_dir = os.path.join(tempfile.mkdtemp(), "sarus_dataspec")
        os.makedirs(self._parquet_dir, exist_ok=True)
        self._unchecked_dataspecs: t.Set[str] = set()

    def set_admin_ds(
        self, source_ds: st.DataSpec, admin_ds: t.Dict[str, t.Any]
    ) -> None:
        """Attach the admin_ds to a source dataspec in a status."""
        assert source_ds.is_source()
        if source_ds.status([ADMIN_DS]) is None:
            stt.ready(
                source_ds,
                task=ADMIN_DS,
                properties={ADMIN_DS: json.dumps(admin_ds)},
            )

    def default_delta(self, dataspec: st.DataSpec) -> t.Optional[float]:
        """Get the default delta of a dataspec."""
        sources_ds = dataspec.sources(sp.type_name(sp.Dataset))
        sources_ds = [
            source for source in sources_ds if not source.is_public()
        ]
        source = sources_ds.pop()
        status = source.status(task_names=[ADMIN_DS])
        if status is None:
            return None
        stage = status.task(task=ADMIN_DS)
        assert stage.ready()
        admin_ds = json.loads(stage[ADMIN_DS])
        # TODO : revise this when the API provides a unified endpoint for
        # applied access rule
        delta = admin_ds.get("accesses").pop(0).get("delta")
        return float(delta)

    def consumed_epsilon(self, dataspec: st.DataSpec) -> float:
        """Get the consumed epsilon from the server."""
        # TODO works only when the value is already computed on the server side
        epsilon = api.get_epsilon(self.client(), dataspec.uuid())
        return epsilon

    def client(self) -> Client:
        """Return the sarus.Client object used to make API calls."""
        return self._client

    def is_cached(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is cached locally."""
        return True

    def is_computation_remote(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is remotely computed."""
        # `select_sql` transforms are delegated, whether mock or not
        if dataspec.is_transformed():
            transform = dataspec.transform()
            if transform.protobuf().spec.HasField(
                "select_sql"
            ) or transform.protobuf().spec.HasField("dp_select_sql"):
                computation_graph = self.computation_graph(dataspec)
                referrables = list(computation_graph["dataspecs"]) + list(
                    computation_graph["transforms"]
                )
                api.push_dataspec_graph(self._client, referrables)

        ds = api.get_dataspec(self.client(), dataspec.uuid())
        return ds is not None

    def is_big_data(self, dataset: st.Dataset) -> bool:
        """Method to compute is_big data.

        Method to compute is_big data if the status
        is not in the storage and the dataset is source
        we retrieve it via an api call
        """
        last_status = self.status(dataset, task_name=BIG_DATA_TASK)
        if last_status:
            stage = last_status.task(task=BIG_DATA_TASK)
            is_big = stage.properties().get(IS_BIG_DATA)
            return is_big == str(True)

        if dataset.is_source():
            status_proto = api.dataspec_status(
                self.client(), dataset.uuid(), task_names=[BIG_DATA_TASK]
            )
            stt.ready(
                dataset,
                manager=self,
                task=BIG_DATA_TASK,
                properties=status_proto.task_stages[BIG_DATA_TASK].properties,
            )
            return status_proto.task_stages[BIG_DATA_TASK].properties[
                IS_BIG_DATA
            ] == str(True)
        return super().is_big_data(dataset)

    def push(self, dataspec: st.DataSpec) -> None:
        """Push a Dataspec's computation graph on the server."""
        computation_graph = self.computation_graph(dataspec)
        referrables = list(computation_graph["dataspecs"]) + list(
            computation_graph["transforms"]
        )
        api.push_dataspec_graph(self.client(), referrables)

    def rewrite(
        self, dataspec: st.DataSpec, target_epsilon: t.Optional[float] = None
    ) -> t.Tuple[st.DataSpec, DataspecPrivacyPolicy]:
        """Rewrite an alternative Dataspec."""
        if target_epsilon is not None:
            default_delta = self.default_delta(dataspec)
            if default_delta is None:
                raise ValueError(f"Default delta not defined for {dataspec}")
            privacy_limit = DeltaEpsilonLimit({default_delta: target_epsilon})
        else:
            privacy_limit = None

        alt_dataspec_uuid, privacy_policy = api.rewrite_dataspec(
            self.client(), dataspec.uuid(), privacy_limit=privacy_limit
        )
        api.pull_dataspec_graph(self.client(), alt_dataspec_uuid)
        alt_dataspec = self.storage().referrable(alt_dataspec_uuid)
        return alt_dataspec, privacy_policy

    def launch(self, dataspec: st.DataSpec) -> None:
        """Launch a Dataspec's computation on the server."""
        api.launch_dataspec(self.client(), dataspec.uuid())

    def push_sql_dataset(self, dataset: st.Dataset):
        api.push_sql_dataset(self.client(), dataset.uuid())

    def mock_value(
        self,
        transform: st.Transform,
        *arguments: t.Union[st.DataSpec, st.Transform],
        **named_arguments: t.Union[st.DataSpec, st.Transform],
    ) -> t.Any:
        value = super().mock_value(transform, *arguments, **named_arguments)
        for argument in (*arguments, *named_arguments.values()):
            self.backpropagate_validation(argument)
        return value

    def type_checked(self, dataspec: st.DataSpec) -> bool:
        """Whether this wrapper was chosen from an hint or running the code."""
        return dataspec.uuid() not in self._unchecked_dataspecs

    def backpropagate_validation(
        self, dataspec: t.Union[st.DataSpec, st.Transform]
    ) -> None:
        def check(ds: st.DataSpec) -> None:
            self._unchecked_dataspecs.remove(ds.uuid())

        def not_type_checked(ds: st.DataSpec) -> bool:
            return isinstance(ds, st.DataSpec) and not self.type_checked(ds)

        class Validator(st.Visitor):
            """Processes only the unchecked dataspecs."""

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                for argument in (*arguments, *named_arguments.values()):
                    if not_type_checked(argument):
                        argument.accept(self)

            def other(self, visited: st.DataSpec) -> None:
                pass

            def all(self, visited: st.DataSpec) -> None:
                check(visited)

        if not_type_checked(dataspec):
            visitor = Validator()
            dataspec.accept(visitor)

    def python_type(self, dataspec: st.DataSpec) -> str:
        """Return the Python class name of a DataSpec.

        This is used to instantiate the correct DataSpecWrapper class.
        """
        python_type_att = dataspec.attribute(name=PYTHON_TYPE)
        if python_type_att is not None:
            return python_type_att.properties().get(PYTHON_TYPE)

        if not dataspec.is_transformed():
            return str(t.Iterator[pa.RecordBatch])

        transform = dataspec.transform()
        if not transform.is_external():
            # TODO: have a python type in SDS for all transforms
            if transform.spec() in ["fit_model", "fit_model_dp"]:
                type_str = str(TrainState)
            else:
                type_str = str(t.Iterator[pa.RecordBatch])
        else:
            ds_args, ds_kwargs = dataspec.parents()
            hint = None
            if not self.client().force_runtime_check:
                try:
                    # allow for SDS without functionality
                    implementation = external_implementation(transform)
                    hint = implementation.py_output_hint(
                        transform, *ds_args, **ds_kwargs
                    )
                    # TODO: add a flag to track that some computations were not run
                except Exception:
                    pass
            if hint is not None:
                type_str = hint
                self._unchecked_dataspecs.add(dataspec.uuid())
            else:
                mock_value = self.mock_value(transform, *ds_args, **ds_kwargs)
                type_str = str(type(mock_value))

        attach_properties(
            dataspec, name=PYTHON_TYPE, properties={PYTHON_TYPE: type_str}
        )
        return type_str

    Edge = t.Tuple[st.DataSpec, st.DataSpec, st.Transform]

    def computation_graph(
        self,
        dataspec: st.DataSpec,
    ) -> t.Dict[str, t.Union[st.DataSpec, st.Transform, st.Attribute, Edge]]:
        """Retreive all items necessary to compute a DataSpec.

        This function is used intensively to post DataSpecs, draw dot
        representationss, fetch statuses, and so on.
        """
        storage = self.storage()

        class ComputationGraphVisitor(st.Visitor, st.TransformVisitor):
            dataspecs: t.List[st.DataSpec] = list()
            transforms: t.Set[st.Transform] = set()
            lambdas: t.Set[st.Transform] = set()
            edges: t.Set[t.Tuple[st.DataSpec, st.DataSpec, st.Transform]] = (
                set()
            )
            attributes: t.Set[st.Attribute] = set()
            variant_constraints: t.Set[st.VariantConstraint] = set()
            graph: t.Dict[str, t.Set[str]] = dict()

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                if visited not in self.dataspecs:
                    self.dataspecs.append(visited)

                    attributes: t.List[st.Attribute] = storage.referring(
                        visited, type_name=sp.type_name(sp.Attribute)
                    )
                    # Don't send MOCK and PYTHON_TYPE attributes to the server
                    self.attributes.update(
                        [
                            att
                            for att in attributes
                            if att.name() not in [MOCK, PYTHON_TYPE]
                        ]
                    )

                    variant_constraints = storage.referring(
                        visited, type_name=sp.type_name(sp.VariantConstraint)
                    )
                    self.variant_constraints.update(
                        [vc for vc in variant_constraints]
                    )

                    self.transforms.add(transform)
                    for argument in arguments:
                        argument.accept(self)
                        self.edges.add((argument, visited, transform))
                        if isinstance(argument, st.Transform):
                            self.lambdas.add(argument)
                    for _, argument in named_arguments.items():
                        argument.accept(self)
                        self.edges.add((argument, visited, transform))
                        if isinstance(argument, st.Transform):
                            self.lambdas.add(argument)

            def other(self, visited: st.DataSpec) -> None:
                if visited not in self.dataspecs:
                    assert isinstance(visited, st.DataSpec)
                    self.dataspecs.append(visited)

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                self.transforms.add(visited)
                self.transforms.add(transform)
                for argument in arguments:
                    argument.accept(self)
                for _, argument in named_arguments.items():
                    argument.accept(self)

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                self.transforms.add(visited)

        visitor = ComputationGraphVisitor()
        dataspec.accept(visitor)

        return {
            "dataspecs": visitor.dataspecs,
            "transforms": visitor.transforms,
            "attributes": visitor.attributes,
            "variant_constraints": visitor.variant_constraints,
            "edges": visitor.edges,
            "lambdas": visitor.lambdas,
        }

    def has_transform(
        self,
        dataspec: st.DataSpec,
        transform_name: str,
    ) -> bool:
        """Check if the computation graph contains some transforms.

        This function is used to check if we have rewritten the computation graph to
        turn big data dataspec into smaller one.
        """

        class HasTransformVisitor(st.Visitor, st.TransformVisitor):
            dataspecs: t.List[st.DataSpec] = list()
            transforms: t.Set[st.Transform] = set()
            has_transform: bool = False
            transform_of_interest: t.Optional[st.Transform] = None

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                if visited not in self.dataspecs:
                    self.dataspecs.append(visited)
                    if transform.name() == transform_name:
                        self.has_transform = True
                        self.transform_of_interest = transform
                    else:
                        self.transforms.add(transform)
                        for argument in arguments:
                            argument.accept(self)
                        for _, argument in named_arguments.items():
                            argument.accept(self)

            def other(self, visited: st.DataSpec) -> None:
                if visited not in self.dataspecs:
                    assert isinstance(visited, st.DataSpec)
                    self.dataspecs.append(visited)

            def composed(
                self,
                visited: st.Transform,
                transform: st.Transform,
                *arguments: st.Transform,
                **named_arguments: st.Transform,
            ) -> None:
                self.transforms.add(visited)
                self.transforms.add(transform)
                for argument in arguments:
                    argument.accept(self)
                for _, argument in named_arguments.items():
                    argument.accept(self)

            def variable(
                self,
                visited: st.Transform,
                name: str,
                position: int,
            ) -> None:
                self.transforms.add(visited)

        visitor = HasTransformVisitor()
        dataspec.accept(visitor)
        return visitor.has_transform, visitor.transform_of_interest

    def _delete_local(self, uuid: str) -> None:
        """Delete a DataSpec locally. MOCKs also have to be deleted."""
        would_delete = self.storage().all_referrings(uuid)
        additional_cleanup = []
        for uuid in would_delete:
            item = self.storage().referrable(uuid)
            if item.prototype() in [sp.Dataset, sp.Scalar]:
                try:
                    mock = item.variant(st.ConstraintKind.MOCK)
                except Exception:
                    pass
                else:
                    if mock:
                        additional_cleanup.append(mock)

        self.storage().delete(uuid)
        for item in additional_cleanup:
            self.storage().delete(item.uuid())

    def dot(
        self,
        dataspec: st.DataSpec,
        symbols: t.Optional[t.Dict[str, t.Optional[str]]] = None,
        policy: t.Optional[DataspecPrivacyPolicy] = None,
        remote: bool = True,
        task_names: t.Optional[t.List[str]] = None,
    ) -> str:
        """Graphviz dot language representation of the graph.

        Statuses are represented with a color code. The symbols are the
        caller's symbol for the DataSpec wrapper
        (see DataSpecWrapper.dataspec_wrapper_symbols).
        """
        if symbols is None:
            symbols = {}
        if task_names is None:
            task_names = [ARROW_TASK, SCALAR_TASK]
        api.pull_dataspec_graph(self.client(), dataspec.uuid())
        graph = self.computation_graph(dataspec)

        # Filter all the dataspecs that are above the source dataset
        source_ds = [
            ds
            for ds in graph["dataspecs"]
            if ds.is_transformed() and ds.transform().spec() == "assign_budget"
        ]
        syn_ds = [
            ds
            for ds in graph["dataspecs"]
            if ds.is_transformed() and ds.transform().spec() == "synthetic"
        ]
        top_ds = syn_ds[0] if syn_ds else source_ds[0]
        onboarded_graph = self.computation_graph(top_ds)

        dataspecs_to_show = (
            set(graph["dataspecs"]) - set(onboarded_graph["dataspecs"])
        ) | {top_ds}
        edges_to_show = set(graph["edges"]) - set(onboarded_graph["edges"])

        TASK = (
            ARROW_TASK if dataspec.prototype() == sp.Dataset else SCALAR_TASK
        )

        # Get statuses wether remote or local
        if remote:
            statuses_proto = api.pull_dataspec_status_graph(
                self._client, dataspec.uuid(), task_names
            )
            statuses = {
                proto.dataspec: stt.Status(proto, store=False)
                for proto in statuses_proto
            }
        else:
            statuses = {
                ds.uuid(): stt.last_status(ds, task=TASK)
                for ds in dataspecs_to_show
            }

        # Transform names to SDK's identifiers
        from sarus.scripts.generate_op_list import op_list

        op_mapping = op_list()

        def transform_label(name: str) -> str:
            """Map the transform name to the SDK's name."""
            if name.endswith("_DP"):
                is_dp = True
                name = name[:-3]
            else:
                is_dp = False

            info = op_mapping[op_mapping.code == name]
            if info.shape[0] == 0:
                if name == "dp_select_sql":
                    return "dp_sql"
                return name
            info = info.iloc[0]
            if pd.isna(info.function):
                label = info.method
            else:
                label = info.function

            if is_dp:
                label = f"dp_{label}"

            return label

        # Add nodes
        edges, nodes, props, lambdas = [], [], [], []
        for ds in dataspecs_to_show:
            status = statuses.get(ds.uuid())
            if ds.uuid() == dataspec.uuid():
                nodes.append(
                    self.dataspec_repr(ds, status, symbols, policy=policy)
                )
            else:
                nodes.append(self.dataspec_repr(ds, status, symbols))

        edges_dict = defaultdict(list)
        for parent, child, transform in edges_to_show:
            tr_label = transform_label(transform.name())
            edges_dict[(child.uuid(), tr_label)].append(parent.uuid())

        for (child, tr_label), parents in edges_dict.items():
            if len(parents) == 1:
                parent = parents.pop()
                edges.append(
                    f'"{parent}" -> "{child}" ' f'[label="{tr_label}"];'
                )
            else:
                # concentrate the edges
                # https://mike42.me/blog/2015-02-how-to-merge-edges-in-graphviz
                intermediate = f"{child}_intermediate"
                parents = map(lambda x: f'"{x}"', parents)
                edges.append(
                    f'"{intermediate}" [shape=point,width=0.01,height=0.01];\n'
                    f'{{{", ".join(parents)}}} -> "{intermediate}" [dir=none];\n'
                    f'"{intermediate}" -> "{child}" [label="{tr_label}"];'
                )

        for i, transform in enumerate(graph["lambdas"]):
            lambdas.append(
                transform.dot().replace(
                    "digraph {",
                    f'subgraph cluster_{i} {{\ncolor=black;\nlabel="Traced function";',
                )
            )

        props = [
            "nodesep=0.1;",
            "ranksep=0.1;",
            'node [style="rounded,filled" fontsize="12" margin="0.055,0.055"]',
            'edge [fontsize="10" fontname="Liberation Mono:style=Regular"]',
        ]
        dot = ["digraph {"] + props + nodes + edges + lambdas + ["}"]
        return "\n".join(dot)

    def dataspec_repr(
        self,
        dataspec: st.DataSpec,
        status: t.Optional[st.Status],
        symbols: t.Dict[str, t.Optional[str]],
        info: t.Literal["status", "privacy"] = "privacy",
        policy: t.Optional[DataspecPrivacyPolicy] = None,
    ) -> str:
        """Style a graph node depending on its status and symbol."""
        # Colors tuple represent (fillcolor, color, fontcolor)
        STATUS_COLORS = {
            "error": ("#ff9c9c", "black", "black"),
            "ready": ("#9cffc5", "black", "black"),
            "pending": ("#ffc89c", "black", "black"),
            "processing": ("#9cbeff", "black", "black"),
            "no_status": ("#ffffff", "#ff9c9c", "black"),
        }
        PRIVACY_COLORS = {
            "PUP": ("#428e92", "#428e92", "white"),
            "DP": ("#006064", "#006064", "white"),
            "NOT_PUP": ("white", "#428e92", "#428e92"),
        }
        if info == "status":
            TASK = (
                ARROW_TASK
                if dataspec.prototype() == sp.Dataset
                else SCALAR_TASK
            )
            stage = status.task(TASK) if status else None
            key = stage.stage() if stage else "no_status"
            fillcolor, color, fontcolor = STATUS_COLORS[key]
        else:
            if dataspec.is_published():
                key = "DP"
            elif dataspec.is_pup():
                key = "PUP"
            else:
                key = "NOT_PUP"

            fillcolor, color, fontcolor = PRIVACY_COLORS[key]

        # Dataspec name
        name = ""
        ds_name = dataspec.name()
        if dataspec.is_transformed():
            transform = dataspec.transform()
            if transform.spec() == "synthetic":
                name = "synthetic_data\n"
            elif transform.spec() == "assign_budget":
                name = "source_data\n"
            elif transform.spec() == "derive_seed":
                name = "local_seed"
            else:
                name = f"{ds_name}\n" if ds_name != "Transformed" else ""
        else:
            if ds_name == "seed":
                name = "master_seed"
            else:
                name = ds_name

        # Class & variable name
        var_name, classname = symbols.get(dataspec.uuid(), ("", ""))
        name = var_name if var_name else name
        classname = f" ({classname})" if classname else classname

        # Status info
        status_msg = ""
        if info == "status":
            if stage:
                status_msg = (
                    stage.properties().get("message", "").replace('"', "'")
                )
            else:
                status_msg = "No status found."
            if status_msg:
                status_msg = "\n" + status_msg

        # Privacy type
        pup_type = " (PUP)" if dataspec.is_pup() else ""

        # Privacy policy
        policy_msg = ""
        if policy:
            policy_msg = {
                DataspecPrivacyPolicy.WHITE_LISTED: "Whitelisted",
                DataspecPrivacyPolicy.DP: "Differentially-Private",
                DataspecPrivacyPolicy.SYNTHETIC: "Synthetic",
                DataspecPrivacyPolicy.PUBLIC: "Public",
            }[policy]
            policy_msg = f"{policy_msg}\n"

        # Assemble
        label = f"{policy_msg}{name}{classname}{pup_type}"
        return (
            f'"{dataspec.uuid()}"[label="{label}", '
            f'fillcolor="{fillcolor}", color="{color}", '
            f'fontcolor="{fontcolor}" shape="box"]'
        )

    async def async_schema_op(self, dataset: st.Dataset) -> st.Schema:
        # quickfix for select sql, to be removed when there is an endpoint
        # for schema
        if dataset.is_transformed() and (
            dataset.transform().spec() == "select_sql"
            or dataset.transform().spec() == "dp_select_sql"
        ):
            return api.pull_dataspec_schema(self._client, dataset.uuid())

        if dataset.is_transformed():
            return await TransformedDataset(dataset).schema()
        else:
            raise ValueError("Dataset should be transformed.")

    async def async_to_arrow_op(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Route a Dataset to its Op implementation.

        When the computation is not delegated the manager should also be
        able to compute the value.
        """
        if dataset.is_transformed():
            iterator = await TransformedDataset(dataset).to_arrow(
                batch_size=batch_size
            )
            return iterator

        else:
            raise ValueError("Dataset should be transformed.")

    async def async_value_op(self, scalar: st.Scalar) -> t.Any:
        """Route a Scalar to its Op implementation.

        This method is shared between API and Worker because when the data is
        not cached the API manager should also be able to compute the value.
        """
        if scalar.is_transformed():
            return await TransformedScalar(scalar).value()
        else:
            return await SourceScalar(scalar).value()

    def copy_status_from_server(
        self, dataspec: st.DataSpec, task_names: t.List[str]
    ) -> None:
        """Method to temporary to retrieve the schemas.

        This method is to be used only temporary to retrieve the
        schemas from the server when a dataset is created.
        """
        status = api.dataspec_status(
            self.client(), dataspec.uuid(), task_names=task_names
        )
        for task_name in task_names:
            if dataspec.status([task_name]) is None:
                stt.ready(
                    dataspec,
                    manager=self,
                    task=task_name,
                    properties=status.task_stages[task_name].properties,
                )

    def dataspec_computation(
        self,
        dataspec: st.DataSpec,
    ) -> BaseComputation:
        """Return the computation for a DataSpec."""
        proto = dataspec.prototype()
        if proto == sp.Dataset:
            return self.to_arrow_computation
        return self.value_computation

    def computation_timeout(self, dataspec: st.DataSpec) -> int:
        return 1800

    def computation_max_delay(self, dataspec: st.DataSpec) -> int:
        return 10


def manager(
    storage: storage_typing.Storage, client, **kwargs: t.Any
) -> SDKManager:
    """Create the SDK manager."""
    properties = {"type": "sdk_manager"}
    properties.update(kwargs)
    return SDKManager(
        storage=storage,
        protobuf=sp.Manager(properties=properties),
        client=client,
    )



from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_property
import numpy as np

try:
    import sklearn.model_selection as sk_model_selection
    from sklearn.model_selection import *  # noqa: F401, F403
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class TimeSeriesSplit(DataSpecWrapper[sk_model_selection.TimeSeriesSplit]):
    @sarus_init("sklearn.SK_TIME_SERIES_SPLIT")
    def __init__(
        self,
        *,
        n_splits=5,
        n_repeats=10,
        random_state=None,
        _dataspec=None,
    ): ...


class RepeatedStratifiedKFold(
    DataSpecWrapper[sk_model_selection.RepeatedStratifiedKFold]
):
    @sarus_init("sklearn.SK_REPEATED_STRATIFIED_KFOLD")
    def __init__(
        self,
        *,
        n_splits=5,
        n_repeats=10,
        random_state=None,
        _dataspec=None,
    ): ...


class KFold(DataSpecWrapper[sk_model_selection.KFold]):
    @sarus_init("sklearn.SK_KFOLD")
    def __init__(self, n_splits=5, *, shuffle=False, random_state=None): ...


class GridSearchCV(DataSpecWrapper[sk_model_selection.GridSearchCV]):
    @sarus_init("sklearn.SK_GRID_SEARCH_CV")
    def __init__(
        self,
        *,
        estimator,
        param_grid,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score=np.nan,
        return_train_score=False,
    ): ...

    @sarus_property("sklearn.SK_BEST_PARAMS_")
    def best_params_(self): ...

    @sarus_property("sklearn.SK_BEST_SCORE_")
    def best_score_(self): ...


register_ops()



import numpy as np

from sarus.dataspec_wrapper import DataSpecWrapper


class Bool(DataSpecWrapper[np.bool_]): ...


class Byte(DataSpecWrapper[np.byte]): ...


class UByte(DataSpecWrapper[np.ubyte]): ...


class Short(DataSpecWrapper[np.short]): ...


class UShort(DataSpecWrapper[np.ushort]): ...


class Intc(DataSpecWrapper[np.intc]): ...


class UIntc(DataSpecWrapper[np.uintc]): ...


class Int_(DataSpecWrapper[np.int_]): ...


class UInt_(DataSpecWrapper[np.uint]): ...


class LongLong(DataSpecWrapper[np.longlong]): ...


class ULongLong(DataSpecWrapper[np.ulonglong]): ...


class Single(DataSpecWrapper[np.single]): ...


class Double(DataSpecWrapper[np.double]): ...


class LongDouble(DataSpecWrapper[np.longdouble]): ...


class CSingle(DataSpecWrapper[np.csingle]): ...


class CDouble(DataSpecWrapper[np.cdouble]): ...


class CLongDouble(DataSpecWrapper[np.clongdouble]): ...



from __future__ import annotations

import logging
import typing as t
from functools import partial

import pandas as pd

from sarus.dataspec_wrapper import (
    IGNORE_WARNING,
    DataSpecVariant,
    DataSpecWrapper,
)
import warnings
from sarus.serialization import trace
from sarus.typing import SPECIAL_WRAPPER_ATTRIBUTES
from sarus.utils import (
    LocalSDKContext,
    create_lambda_op,
    create_method,
    get_parent_module_name,
    global_context,
    register_ops,
    sarus_custom_method,
    sarus_init,
    sarus_method,
    sarus_property,
    serialize_external,
)

logger = logging.getLogger(__name__)

sarus_method = partial(sarus_method, use_parent_module=True)
sarus_property = partial(sarus_property, use_parent_module=True)
sarus_init = partial(sarus_init, use_parent_module=True)


class DataFrame(DataSpecWrapper[pd.DataFrame]):
    @sarus_init("pandas.PD_DATAFRAME")
    def __init__(
        self, data=None, index=None, columns=None, dtype=None, copy=None
    ): ...

    @sarus_method("std.SETITEM", inplace=True)
    def __setitem__(self, key, newvalue): ...

    @sarus_property("pandas.PD_COLUMNS")
    def columns(self): ...

    @sarus_property("pandas.PD_AXES")
    def axes(self): ...

    @sarus_property("pandas.PD_INDEX")
    def index(self): ...

    @sarus_property("pandas.PD_SHAPE")
    def shape(self): ...

    @sarus_property("pandas.PD_NDIM")
    def ndim(self): ...

    @sarus_property("pandas.PD_SIZE")
    def size(self): ...

    @sarus_property("pandas.PD_VALUES")
    def values(self): ...

    @property
    def loc(self) -> _SarusLocIndexer:
        return _SarusLocIndexer(self)

    @property
    def iloc(self) -> _SarusLocIndexer:
        return _SarusILocIndexer(self)

    def __getattr__(self, name: str) -> t.Any:
        # Overload __getattr__ to enable indexing by column name
        if name in IGNORE_WARNING:
            return object.__getattribute__(self, name)

        if name in SPECIAL_WRAPPER_ATTRIBUTES:
            return super().__getattr__(name=name)

        if name in self.__sarus_eval__(verbose=0).columns:
            return self.loc[:, name]

        return super().__getattr__(name=name)

    def __setattr__(self, name: str, value: t.Any) -> None:
        # Overload __setattr__ to enable setting by column name
        if name in SPECIAL_WRAPPER_ATTRIBUTES:
            return super().__setattr__(name, value)

        if name in self.__sarus_eval__().columns:
            key = (slice(None, None), name)
            new_df = _pd_set_loc(self, key, value)
            self._set_dataspec(new_df._dataspec)
            return

        return super().__setattr__(name, value)

    def copy(self, deep: bool = False) -> DataFrame:
        return DataFrame.from_dataspec(
            self.dataspec(kind=DataSpecVariant.USER_DEFINED)
        )

    @sarus_custom_method("pandas.PD_APPLY")
    def apply(
        self,
        func: t.Callable,
        axis: int = 0,
        raw: bool = False,
        result_type: t.Optional[
            t.Literal["expand", "reduce", "broadcast"]
        ] = None,
        args: t.Tuple = (),
        by_row: t.Union[t.Literal["compat"], False] = False,
    ):
        warnings.warn(
            "Apply on a dataframe only works if"
            "all columns have the same dtype"
        )
        # Trace function
        if axis == 0:
            example = self[self.columns[0]]
        else:
            example = self.iloc[0]

        traced_func = trace(func)(example)

        new_dataspec, traced_transform = serialize_external(
            "pandas.PD_APPLY",
            self,
            func=traced_func,
            axis=axis,
            raw=raw,
            result_type=result_type,
            args=args,
            by_row=by_row,
        )
        context: LocalSDKContext = global_context()

        new_dataspec_wrapper = context.wrapper_factory().create(new_dataspec)
        new_dataspec_wrapper._set_traced_transform(traced_transform)
        return new_dataspec_wrapper

    @sarus_custom_method("pandas.PD_APPLYMAP")
    def applymap(
        self,
        func: t.Callable,
        na_action: t.Optional[t.Literal["ignore"]] = None,
    ):
        # Trace function
        example = self.iloc[0][0]
        traced_func = trace(func)(example)

        new_dataspec, traced_transform = serialize_external(
            "pandas.PD_APPLYMAP",
            self,
            func=traced_func,
            na_action=na_action,
        )
        context: LocalSDKContext = global_context()

        new_dataspec_wrapper = context.wrapper_factory().create(new_dataspec)
        new_dataspec_wrapper._set_traced_transform(traced_transform)
        return new_dataspec_wrapper


# This registration process is an exception
_pd_loc = create_method(
    "pandas.PD_LOC",
    module_name=__name__,
    class_name="DataFrame",
    method_name="loc",
    doc_module_name=get_parent_module_name(__name__),
)
_pd_iloc = create_method(
    "pandas.PD_ILOC",
    module_name=__name__,
    class_name="DataFrame",
    method_name="iloc",
    doc_module_name=get_parent_module_name(__name__),
)
_ = create_method(
    "pandas.PD_LOC",
    module_name=__name__,
    class_name="Series",
    method_name="loc",
    doc_module_name=get_parent_module_name(__name__),
)
_ = create_method(
    "pandas.PD_ILOC",
    module_name=__name__,
    class_name="Series",
    method_name="iloc",
    doc_module_name=get_parent_module_name(__name__),
)
_pd_set_loc = create_lambda_op("pandas.PD_SET_LOC")
_pd_set_iloc = create_lambda_op("pandas.PD_SET_ILOC")


class _SarusLocIndexer:
    def __init__(self, df: DataFrame) -> None:
        self.df = df

    def __getitem__(self, key) -> DataFrame:
        return _pd_loc(self.df, key)

    def __setitem__(self, key, newvalue) -> None:
        new_df = _pd_set_loc(self.df, key, newvalue)
        self.df._set_dataspec(new_df._dataspec)


class _SarusILocIndexer:
    def __init__(self, df: DataFrame) -> None:
        self.df = df

    def __getitem__(self, key) -> DataFrame:
        return _pd_iloc(self.df, key)

    def __setitem__(self, key, newvalue) -> None:
        new_df = _pd_set_iloc(self.df, key, newvalue)
        self.df._set_dataspec(new_df._dataspec)


class Series(DataSpecWrapper[pd.Series]):
    @sarus_init("pandas.PD_SERIES")
    def __init__(
        data=None,
        index=None,
        dtype=None,
        name=None,
        copy=False,
        fastpath=False,
    ): ...

    @sarus_method("std.SETITEM", inplace=True)
    def __setitem__(self, key, newvalue): ...

    @sarus_property("pandas.PD_INDEX")
    def index(self): ...

    @sarus_property("pandas.PD_DTYPE")
    def dtype(self): ...

    @sarus_property("pandas.PD_SHAPE")
    def shape(self): ...

    @sarus_property("pandas.PD_NDIM")
    def ndim(self): ...

    @sarus_property("pandas.PD_SIZE")
    def size(self): ...

    @sarus_property("pandas.PD_NAME")
    def name(self): ...

    @sarus_property("pandas.PD_VALUES")
    def values(self): ...

    def copy(self, deep: bool = False) -> Series:
        return Series.from_dataspec(
            self.dataspec(kind=DataSpecVariant.USER_DEFINED)
        )

    @sarus_custom_method("pandas.PD_APPLY")
    def apply(
        self,
        func: t.Callable,
        convert_dtype: bool = True,
        args: t.Tuple = (),
        by_row: t.Union[False, t.Literal["compat"]] = False,
    ):
        # Trace function
        example = self[0]
        traced_func = trace(func)(example)

        new_dataspec, traced_transform = serialize_external(
            "pandas.PD_APPLY",
            self,
            func=traced_func,
            convert_dtype=convert_dtype,
            args=args,
            by_row=by_row,
        )
        context: LocalSDKContext = global_context()

        new_dataspec_wrapper = context.wrapper_factory().create(new_dataspec)
        new_dataspec_wrapper._set_traced_transform(traced_transform)
        return new_dataspec_wrapper


register_ops(use_parent_module=True)



"""Sarus Skopt package documentation."""

# flake8: noqa
try:
    from skopt import *
except ModuleNotFoundError:
    pass

try:
    import sarus.skopt.searchcv as searchcv
    from sarus.skopt.searchcv import BayesSearchCV
except NameError:
    pass

__all__ = ["BayesSearchCV"]



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import imblearn.pipeline as imb_pipeline
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class Pipeline(DataSpecWrapper[imb_pipeline.Pipeline]):
    @sarus_init("imblearn.IMB_PIPELINE")
    def __init__(steps, *, memory=None, verbose=False): ...

    @sarus_method("sklearn.SK_PIPELINE_FIT", inplace=True)
    def fit(self, X, y=None): ...


register_ops()



"""Sarus Shap package documentation."""

# flake8: noqa
from sarus.utils import register_ops

try:
    from shap import *

    from .explainers import Additive as AdditiveExplainer
    from .explainers import *
    from .explainers import Explainer
    from .explainers import GPUTree as GPUTreeExplainer
    from .explainers import Kernel as KernelExplainer
    from .explainers import Linear as LinearExplainer
    from .explainers import Partition as PartitionExplainer
    from .explainers import Permutation as PermutationExplainer
    from .explainers import Sampling as SamplingExplainer
    from .explainers import Tree as TreeExplainer
    from .Explanation import *
    from .maskers import *
    from .plots import *
    from .utils import *
except ModuleNotFoundError:
    bar_plot = None
    summary_plot = None
    decision_plot = None
    multioutput_decision_plot = None
    embedding_plot = None
    force_plot = None
    save_html = None
    group_difference_plot = None
    image_plot = None
    monitoring_plot = None
    partial_dependence_plot = None
    dependence_plot = None
    text_plot = None
    waterfall_plot = None
    initjs = None


register_ops()



from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init

try:
    import imblearn.under_sampling as imb_under_sampling
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class RandomUnderSampler(
    DataSpecWrapper[imb_under_sampling.RandomUnderSampler]
):
    @sarus_init("imblearn.IMB_RANDOM_UNDER_SAMPLER")
    def __init__(
        self, *, sampling_strategy="auto", random_state=None, replacement=False
    ): ...


register_ops()



import logging
import time
import traceback
import typing as t

import pyarrow as pa
import pyarrow.parquet as pq
import sarus_data_spec.status as stt
from sarus_data_spec import typing as st
from sarus_data_spec.constants import ARROW_TASK
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.async_utils import (
    async_iter,
)
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.computations.base import (
    ErrorCatchingAsyncIterator,
)
from sarus_data_spec.manager.computations.local.base import LocalComputation

from sarus.manager.arrow_remote import ToArrowComputationOnServer
from sarus.manager.parquet_local import ToParquetComputation

logger = logging.getLogger(__name__)

# TODO: in SDS some of the default batch_size is 1 and it slows down the SDK by dozen of secondes for large tables
# to remove when the default batch_size is changes in SDS.
BATCH_SIZE = 10000


class ToArrowComputation(LocalComputation[t.AsyncIterator[pa.RecordBatch]]):
    task_name = ARROW_TASK

    def __init__(
        self,
        computing_manager: Base,
        parquet_computation: ToParquetComputation,
        arrow_remote: ToArrowComputationOnServer,
    ) -> None:
        super().__init__(computing_manager)
        self.parquet_computation = parquet_computation
        self.remote_arrow = arrow_remote

    async def prepare(self, dataspec: st.DataSpec) -> None:
        try:
            logger.debug(f"STARTED LOCAL ARROW {dataspec.uuid()}")
            start = time.perf_counter()
            # Only prepare parents since calling `to_arrow` will require the
            # computation of the scalars in the ancestry.
            dataset = t.cast(st.Dataset, dataspec)

            if self.computing_manager().is_cached(dataspec):
                await self.parquet_computation.complete_task(dataspec)

            elif self.computing_manager().is_computation_remote(dataspec):
                await self.arrow_remote.complete_task(dataspec)

            else:
                await self.computing_manager().async_prepare_parents(dataset)
                await self.computing_manager().async_schema(dataset)
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))
        else:
            end = time.perf_counter()
            logger.debug(
                f"FINISHED LOCAL ARROW {dataspec.uuid()} ({end-start:.2f}s)"
            )
            stt.ready(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
            )

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Returns the iterator."""
        if self.computing_manager().is_cached(dataspec):
            status = self.parquet_computation.status(dataspec)
            assert status
            stage = status.task(self.parquet_computation.task_name)
            assert stage
            assert stage.ready()
            cache_path = (
                await self.parquet_computation.result_from_stage_properties(
                    dataspec, stage.properties()
                )
            )
            try:
                ait = async_iter(
                    pq.read_table(source=cache_path).to_batches(
                        max_chunksize=BATCH_SIZE
                    )
                )
            except Exception as e:
                stt.error(
                    dataspec=dataspec,
                    manager=self.computing_manager(),
                    task=self.task_name,
                    properties={
                        "message": traceback.format_exc(),
                        "relaunch": str(True),
                    },
                )
                stt.error(
                    dataspec=dataspec,
                    manager=self.computing_manager(),
                    task=self.parquet_computation.task_name,
                    properties={
                        "message": traceback.format_exc(),
                        "relaunch": str(True),
                    },
                )
                raise stt.DataSpecErrorStatus(
                    (True, traceback.format_exc())
                ) from e
        elif self.is_computation_remote(dataspec):
            status = self.remote_arrow.status(dataspec)
            ait = await self.remote_arrow.result_from_stage_properties(
                dataspec,
                batch_size=BATCH_SIZE,
                properties=status.task(
                    self.remote_arrow.task_name
                ).properties(),
            )
        else:
            ait = await self.computing_manager().async_to_arrow_op(
                dataset=t.cast(Dataset, dataspec), batch_size=BATCH_SIZE
            )
        return ErrorCatchingAsyncIterator(ait, dataspec, self)



from typing import cast

import sarus_data_spec.protobuf as sp
from sarus_data_spec.attribute import Attribute
from sarus_data_spec.context.public import Public as PublicContext
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.factory import Factory
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import Schema
from sarus_data_spec.storage.local import Storage
from sarus_data_spec.transform import Transform
from sarus_data_spec.variant_constraint import VariantConstraint
from sarus_data_spec.type import Type

from sarus.dataspec_wrapper import MetaWrapper
from sarus.manager.sdk_manager import SDKManager, manager
from sarus.typing import Client
from sarus.wrapper_factory import DataSpecWrapperFactory

from ..typing import SyncPolicy


class LocalSDKContext(PublicContext):
    """A default context."""

    def __init__(self, client: Client, verbose: int = 1) -> None:
        super().__init__()
        self._storage = Storage()  # type:ignore
        self._sync_policy = SyncPolicy.SEND_ON_INIT
        self.client = client
        self._manager = manager(self.storage(), self.client)
        self._verbose = verbose

        self._dataspec_factory = Factory()
        self.factory().register(
            sp.type_name(sp.Dataset),
            lambda protobuf, store: Dataset(cast(sp.Dataset, protobuf), store),
        )
        self.factory().register(
            sp.type_name(sp.Scalar),
            lambda protobuf, store: Scalar(cast(sp.Scalar, protobuf), store),
        )
        self.factory().register(
            sp.type_name(sp.Transform),
            lambda protobuf, store: Transform(
                cast(sp.Transform, protobuf), store
            ),
        )
        self.factory().register(
            sp.type_name(sp.Schema),
            lambda protobuf, store: Schema(cast(sp.Schema, protobuf), store),
        )
        self.factory().register(
            sp.type_name(sp.Attribute),
            lambda protobuf, store: Attribute(
                cast(sp.Attribute, protobuf), store
            ),
        )

        self.factory().register(
            sp.type_name(sp.VariantConstraint),
            lambda protobuf, store: VariantConstraint(
                cast(sp.VariantConstraint, protobuf), store
            ),
        )

        self.factory().register(
            sp.type_name(sp.Type),
            lambda protobuf: Type(cast(sp.Type, protobuf)),
        )

        self._wrapper_factory = DataSpecWrapperFactory()
        for (
            python_classname,
            sarus_wrapper_class,
        ) in MetaWrapper._wrapper_classes:
            self._wrapper_factory.register(
                python_classname=python_classname,
                sarus_wrapper_class=sarus_wrapper_class,
            )

    def verbose(self) -> int:
        return self._verbose

    def factory(self) -> Factory:
        return self._dataspec_factory

    def wrapper_factory(self) -> DataSpecWrapperFactory:
        return self._wrapper_factory

    def storage(self) -> Storage:
        return self._storage

    def manager(self) -> SDKManager:
        return self._manager

    def set_sync_policy(self, policy: SyncPolicy) -> None:
        self._sync_policy = policy

    def sync_policy(self) -> SyncPolicy:
        return self._sync_policy



import logging
import os
import pickle as pkl
import time
import traceback
import typing as t

import sarus_data_spec.protobuf as sp
from sarus_data_spec import typing as st
from sarus_data_spec.constants import (
    CACHE_PATH,
    CACHE_PROTO,
    CACHE_SCALAR_TASK,
    CACHE_TYPE,
    ScalarCaching,
)
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.computations.local.base import LocalComputation
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.status import DataSpecErrorStatus, error, ready

from sarus.manager.value_remote import ValueComputationOnServer

logger = logging.getLogger(__name__)


class CacheScalarComputation(LocalComputation[t.Tuple[str, str]]):
    """Class responsible for handling the caching in of a scalar. It wraps a ValueComputation to get the value."""

    task_name = CACHE_SCALAR_TASK

    def __init__(
        self, computing_manager: Base, remote_scalar: ValueComputationOnServer
    ) -> None:
        super().__init__(computing_manager)
        self.remote_scalar = remote_scalar

    async def prepare(self, dataspec: st.DataSpec) -> None:
        logger.debug(f"STARTING CACHE_SCALAR {dataspec.uuid()}")
        start = time.perf_counter()
        try:
            if self.computing_manager().is_computation_remote(dataspec):
                value = await self.remote_scalar.task_result(dataspec)
            else:
                value = await self.computing_manager().async_value_op(
                    scalar=t.cast(Scalar, dataspec)
                )

            if isinstance(value, st.HasProtobuf):
                properties = {
                    CACHE_PROTO: sp.to_base64(value.protobuf()),
                    CACHE_TYPE: sp.type_name(value.prototype()),
                }
            else:
                properties = {
                    CACHE_TYPE: ScalarCaching.PICKLE.value,
                    CACHE_PATH: self.cache_path(dataspec),
                }
                with open(self.cache_path(dataspec), "wb") as f:
                    pkl.dump(value, f)

        except DataSpecErrorStatus as exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(exception.relaunch),
                },
            )
            raise

        except Exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(False),
                },
            )

            raise DataSpecErrorStatus((False, traceback.format_exc()))
        else:
            end = time.perf_counter()
            logger.debug(
                f"FINISHED CACHE_SCALAR {dataspec.uuid()} ({end-start:.2f}s)"
            )
            ready(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties=properties,
            )

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> t.Tuple[str, str]:
        """Reads the cache and returns the value."""
        if properties[CACHE_TYPE] == ScalarCaching.PICKLE.value:
            return properties[CACHE_TYPE], properties[CACHE_PATH]
        return properties[CACHE_TYPE], properties[CACHE_PROTO]

    def cache_path(self, dataspec: st.DataSpec) -> str:
        """Returns the path where to cache the scalar."""
        return os.path.join(
            dataspec.manager().parquet_dir(), f"{dataspec.uuid()}.pkl"
        )



from __future__ import annotations

import logging
from typing import Dict, List, Optional, Union

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import sarus_init, sarus_property

try:
    import numpy.typing as npt
    import shap
    from scipy.sparse import spmatrix
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


class Explanation(DataSpecWrapper[shap.Explanation]):
    @sarus_init("shap.SHAP_EXPLANATION")
    def __init__(
        self,
        values: Union[npt.ArrayLike, spmatrix],
        base_values: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        data: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        display_data: Optional[Dict[str, npt.ArrayLike]] = None,
        instance_names: Optional[List[str]] = None,
        feature_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        output_indexes: Optional[List[int]] = None,
        lower_bounds: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        upper_bounds: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        error_std: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        main_effects: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        hierarchical_values: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        clustering: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        compute_time: Optional[float] = None,
    ): ...

    @sarus_property("shap.SHAP_VALUES")
    def values(self): ...

    @sarus_property("shap.SHAP_SUM")
    def sum(self): ...



import enum
from typing import Literal, Optional, Union, cast

import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
from sarus_data_spec.dataset import make_prompts
from sarus_data_spec.scalar import pretrained_model
from sarus_data_spec.transform import fit_model, generate_from_model

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.pandas.dataframe import DataFrame
from sarus.utils import register_ops


class TrainState: ...


class SampleType(str, enum.Enum):
    PRETRAIN = "pretrain"
    INSTRUCT = "instruct"


class LLM(DataSpecWrapper[TrainState]):
    def __init__(self, name: str, sample_type: SampleType) -> None:
        self.name = name
        if name not in [
            "llama3-8b",
            "open_mistral_7b",
            "llama3-70b",
            "phi_3_mini",
        ]:
            raise ValueError("Unknown model name.")
        model_dataspec = pretrained_model(self.name, sample_type)
        self._set_dataspec(model_dataspec)

    def fit(
        self,
        X: DataSpecWrapper,
        batch_size: int,
        text_field: Optional[str] = None,
        answer_field: Optional[str] = None,
        question_field: Optional[str] = None,
        epochs: int = -1,
        X_val: Optional[DataSpecWrapper] = None,
    ) -> None:
        if not isinstance(X, DataSpecWrapper):
            raise TypeError(
                f"Cannot fit {self.name} on `X` because it is not a Sarus object."
            )

        input_ds = X.dataspec()
        if input_ds.prototype() != sp.Dataset:
            raise TypeError(
                f"Cannot fit {self.name} on `X` because it is not a Sarus Dataset."
            )
        model_ds = self.dataspec()
        if X_val is not None:
            if not isinstance(X_val, DataSpecWrapper):
                raise TypeError("X_val should be a DataSpecWrapper.")
            val_ds = X_val.dataspec()
            fitted_dataspec = fit_model(
                epochs=epochs,
                batch_size=batch_size,
                text_field=text_field,
                question_field=question_field,
                answer_field=answer_field,
            )(model=model_ds, dataset=input_ds, validation_dataset=val_ds)
        else:
            fitted_dataspec = fit_model(
                epochs=epochs,
                batch_size=batch_size,
                text_field=text_field,
                question_field=question_field,
                answer_field=answer_field,
            )(model=model_ds, dataset=input_ds)

        self._set_dataspec(fitted_dataspec)
        return self

    def sample(
        self,
        prompts,
        temperature: float = 1.0,
        max_new_tokens: int = 20,
    ):
        if not isinstance(prompts, st.DataSpec):
            prompts = make_prompts(prompts)
        samples = generate_from_model(
            temperature=temperature, max_new_tokens=max_new_tokens
        )(model=self.dataspec(), prompts=prompts)

        return DataFrame.from_dataspec(samples)

    def __sarus_eval__(
        self,
        target_epsilon: Union[Optional[float], Literal["unlimited"]] = None,
        verbose: Optional[int] = None,
    ) -> str:
        _ = super().__sarus_eval__(target_epsilon, verbose)

        scalar = cast(st.Scalar, self.dataspec())
        if scalar.is_pretrained_model():
            return "model (pretrained)"
        elif scalar.is_fitted_model():
            return "model (fitted)"
        else:
            raise TypeError(
                "The wrapped scalar is not a pretrained or a fitted model."
            )


register_ops()



import inspect
import logging
import os
from datetime import date, datetime, time, timedelta, timezone
from functools import partial, wraps
from typing import Any, Callable, Dict, Optional, Type, Union

import numpy as np
import pandas as pd
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
import sarus_data_spec.status as stt
import yaml
from sarus_data_spec.context import global_context
from sarus_data_spec.status import DataSpecErrorStatus
from sarus_data_spec.transform import external, error_estimation, push_sql

from sarus_data_spec.constants import (
    SARUS_DEFAULT_OUTPUT,
)
from sarus_data_spec.dataset import transformed
from sarus_data_spec.constants import PUSH_SQL_TASK
from sarus_data_spec.dataspec_validator.typing import DataspecPrivacyPolicy
import sarus.manager.dataspec_api as api

from .context.typing import LocalSDKContext
from .typing import (
    DataSpecVariant,
    DataSpecWrapper,
    FederatedWrapper,
    TracedFunction,
)

logger = logging.getLogger(__name__)

config_file = os.path.join(os.path.dirname(__file__), "config.yaml")
with open(config_file) as f:
    config = yaml.load(f.read(), Loader=yaml.Loader)


def module_config(module_name: str) -> Optional[Dict[str, Any]]:
    """Fetch the module's configuration from the config dict."""
    keys = module_name.split(".")
    module_conf = config
    for key in keys:
        if module_conf is None:
            return
        module_conf = module_conf.get(key)
    return module_conf


def eval(
    x: Any,
    target_epsilon: Optional[float] = None,
    verbose: Optional[int] = None,
) -> st.DataSpecValue:
    """Recursively evaluates DataSpecWrappers to values."""
    if isinstance(x, (DataSpecWrapper, FederatedWrapper)):
        return x.__sarus_eval__(target_epsilon, verbose)

    if target_epsilon is not None:
        logger.warning(
            "Ignoring `target_epsilon` since the evaluated object"
            " is not a Sarus object."
        )

    if isinstance(x, list):
        return [eval(x_) for x_ in x]
    elif isinstance(x, tuple):
        return tuple([eval(x_) for x_ in x])
    elif isinstance(x, dict):
        return {eval(k): eval(v) for k, v in x.items()}
    else:
        return x


def eval_perturbation(
    x: Any,
    target_epsilon: Optional[float] = None,
    verbose: Optional[int] = None,
) -> st.DataSpecValue:
    """Evaluates the error due to DP.

    To do so, compute an estimation of the 95th percentile of the absolute
    distance between the true value and DP values.
    If target_epsilon is None, defaults to the same as eval().
    """
    if isinstance(x, DataSpecWrapper):
        manager = x.manager()

        # Dataspec can be a Scalar or a Dataset. If a Dataset, it should be
        # convertible to a pd.Series.
        dataspec = x.dataspec()

        # needed to create new dataspecs with a different name
        # for the rewriter can associate a different seed
        dataspec_transform = dataspec.transform()
        dataspec_type = dataspec.prototype()
        args, kwargs = dataspec.parents()

        # get DP variants
        dp_dataspecs = []
        for i in range(19):
            # send to the server but don't evaluate the computational graph
            # to be sure to have always a different computational graph we
            # can add very small noise to the target epsilon.
            new_ds = transformed(
                dataspec_transform,
                *args,
                dataspec_type=sp.type_name(dataspec_type),
                dataspec_name=f"Perturbation_{i}",
                **kwargs,
            )
            manager.push(new_ds)
            dp_dataspec, _ = manager.rewrite(new_ds, target_epsilon)

            if i == 0 and not dp_dataspec.is_published():
                raise DataSpecErrorStatus(
                    (
                        False,
                        "The dataspec could not be rewritten in DP.",
                    )
                )
            dp_dataspecs.append(dp_dataspec)

        # normally the ErrorEstimation op is whitelisted
        dataspec = error_estimation()(dataspec, *dp_dataspecs)
        manager.push(dataspec)
        alt_dataspec, _ = manager.rewrite(dataspec)
        return alt_dataspec.value()

    if isinstance(x, list):
        # same as eval, but why don't we pass target_epsilon and verbose ?
        return [eval_perturbation(x_) for x_ in x]
    elif isinstance(x, tuple):
        return tuple([eval_perturbation(x_) for x_ in x])
    elif isinstance(x, dict):
        return {
            eval_perturbation(k): eval_perturbation(v) for k, v in x.items()
        }
    else:
        return x


def find_id_by_name_datacon(dataconnections, datacon_name):
    for data in dataconnections:
        if data.get("name") == datacon_name:
            return data.get("id")
    return None


def push_to_table(
    wrapper: DataSpecWrapper,
    data_connection_name: str,
    table: str,
    return_uuid=False,
):
    """Pushes a dataspec to a specified table within a data connection.

    This method constructs the URI for the dataspec, prepares the dataspec for pushing by removing columns not present in the table, and performs the push
    operation using the provided wrapper's manager and client API.

    Args:
        wrapper (DataSpecWrapper): The dataspec wrapper containing the dataspec to be pushed.
        data_connection_name (str): The name of the data connection where the dataspec will be pushed.
        table (str): The name of the table where the dataspec will be pushed, in the format 'schema_name.table_name'.
        return_uuid (bool): If True, the function returns the UUID of the pushed dataspec.

    Returns:
        str: The UUID of the pushed dataspec, if return_uuid is True.

    Raises:
        ValueError: If 'table' does not follow the 'schema_name.table_name' format or if it contains empty values, or if the wrapper type is not supported.
    """
    manager = wrapper.manager()
    client = manager.client()

    # get data_connection_name and table_name
    parts = table.split(".")
    if len(parts) == 2 and all(parts):
        schema_name, table_name = parts
    else:
        raise ValueError(
            "Error: Input must be in the format 'dataconnection_name.schema_name.table_name' with non-empty dataconnection_name, schema_name or table_name."
        )

    dataconnections = client._dataconnections()
    dataconnection_id = find_id_by_name_datacon(
        dataconnections, data_connection_name
    )
    uri = f"sarus://{dataconnection_id}"

    from sarus.sarus import DataFrame, Dataset

    if isinstance(wrapper, Dataset):
        wrapper = wrapper.as_pandas()
    elif isinstance(wrapper, DataFrame):
        pass
    else:
        raise ValueError(
            "Error: Input must be either a {DataFrame} or a {Dataset}. to be pushed"
        )

    try:
        destination_table_columns = (
            client._get_type_destination_table(data_connection_name, table)
            .children()
            .keys()
        )
        wrapper_columns = set(eval(wrapper.columns, verbose=0))
        common_columns = wrapper_columns.intersection(
            destination_table_columns
        )

        if not common_columns:
            raise ValueError(
                f"No common columns with the destination table '{table}'."
            )

        difference_columns = wrapper_columns - common_columns
        if difference_columns:
            logger.warning(
                f"Dropping columns {difference_columns} not present in destination table '{table}'."
            )

        wrapper = wrapper[list(common_columns)]
    except Exception:
        pass

    dataspec = wrapper.dataspec()
    to_be_pushed_dataspec = push_sql(
        data_connection_name, schema_name, table_name, uri
    )(dataspec)
    manager.push(to_be_pushed_dataspec)
    api.push_sql_dataset(client, to_be_pushed_dataspec.uuid())
    print(DataspecPrivacyPolicy.WHITE_LISTED.value)
    print(
        f"Check the evolution of the push to table with sarus.check_push_to_table_status('{to_be_pushed_dataspec.uuid()}')"
    )
    if return_uuid:
        return to_be_pushed_dataspec.uuid()


def check_push_to_table_status(uuid: str, return_status=False):
    """Returns/Prints the status of the computation to sql for the input uuid."""
    context: LocalSDKContext = global_context()
    client = context.manager().client()

    status_proto = api.dataspec_status(
        client=client,
        uuid=uuid,
        task_names=[PUSH_SQL_TASK],
    )
    if status_proto is None:
        return None
    else:
        status = stt.Status(protobuf=status_proto, store=False)
        stage = status.task(PUSH_SQL_TASK).stage()
        print(f"The push to table is in stage {stage}")
        print(status.task(PUSH_SQL_TASK).properties()["message"])
        if return_status:
            return status.task(PUSH_SQL_TASK)


def save(dw: DataSpecWrapper, path: str) -> None:
    """Save a DataSpecWrapper for loading in another notebook."""
    dw.sarus_save(path)


def is_standard_type_iterable(iterable) -> bool:
    """Checks if all elements in an iterable are instances of a specific set of standard types.

    This function is designed to quickly verify if a given iterable (such as list, set, tuple)
    contains only elements of certain standard types (like datetime, string, integer etc.).
    It provides a first faster alternative to checking each item individually for more complex
    types such as DataSpecWrapper or TracedFunction.

    Args:
        iterable (iterable): The iterable to check. This can be a list, set, tuple etc.

    Returns:
        bool: True if all elements in the iterable are of the specified types, False otherwise.
    """
    types = (
        pd.Timestamp,
        datetime,
        timedelta,
        timezone,
        time,
        date,
        np.generic,
        pd.Period,
        pd.Timedelta,
        pd.Interval,
        type,
        pd.api.extensions.ExtensionDtype,
        str,
        int,
        float,
        bool,
        type(None),
    )
    return all(isinstance(x, types) for x in iterable)


def convert_container(
    x: Union[DataSpecWrapper, Any],
) -> Union[DataSpecWrapper, Any]:
    """Recursively convert containers in DataSpecWrappers if one element is a DataSpecWrapper."""
    from sarus.std import Dict, List, Set, Slice, Tuple

    sarus_types = (DataSpecWrapper, TracedFunction)
    common_container_types = (list, set, tuple, dict, slice)
    iterable_container_types = (list, set, tuple)

    if isinstance(x, sarus_types):
        return x
    elif isinstance(x, common_container_types):
        # a first check that is faster than checking DataSpecWrapper or TracedFunction
        if isinstance(
            x, iterable_container_types
        ) and is_standard_type_iterable(x):
            return x

        if isinstance(x, list):
            elems = [convert_container(e) for e in x]
            if any(isinstance(e, sarus_types) for e in elems):
                return List(*elems)
            else:
                return x
        elif isinstance(x, set):
            elems = [convert_container(e) for e in x]
            if any(isinstance(e, sarus_types) for e in elems):
                return Set(*elems)
            else:
                return x
        elif isinstance(x, tuple):
            elems = [convert_container(e) for e in x]
            if any(isinstance(e, sarus_types) for e in elems):
                return Tuple(*elems)
            else:
                return x
        elif isinstance(x, dict):
            if is_standard_type_iterable(x.values()):
                return x

            elems = {k: convert_container(v) for k, v in x.items()}
            if any(isinstance(e, sarus_types) for e in elems.values()):
                return Dict(**elems)
            else:
                return x
        elif isinstance(x, slice):
            elems = [
                x.start,
                x.stop,
                x.step,
            ]
            if any([isinstance(e, sarus_types) for e in elems]):
                return Slice(*elems)
            else:
                return x
    else:
        return eval(x)


def eval_policy(x: Any) -> Optional[str]:
    """The alternative dataspec's privacy policy."""
    if isinstance(x, DataSpecWrapper):
        return x.__eval_policy__()
    else:
        return None


_registered_methods = []
_registered_functions = []


class register_method:
    """This decorator method allows to register methods declared in classes.

    It uses this behavior since Python 3.6
    https://docs.python.org/3/reference/datamodel.html#object.__set_name__
    """

    def __init__(
        self, method: Callable, code_name: str, use_parent_module: bool = False
    ) -> None:
        self.method = method
        self.code_name = code_name
        self.use_parent_module = use_parent_module

    def __set_name__(self, owner: Type, name: str) -> None:
        global _registered_methods
        module_name = owner.__module__
        if self.use_parent_module:
            doc_module_name = get_parent_module_name(module_name)
        else:
            doc_module_name = module_name
        _registered_methods.append(
            (
                doc_module_name,
                owner.__name__,
                name,
                self.code_name,
            )
        )
        setattr(owner, name, self.method)


def get_parent_module_name(module_name):
    return ".".join(module_name.split(".")[:-1])


def register_ops(use_parent_module: bool = False):
    """Monkey-patching standard libraries to have Sarus functions.

    This functions is intended to be called in a Sarus module. The module's
    local variables will be modified dynamically (monkey patching) to replace
    some functions or methods by Sarus equivalent operations.

    Technically, we get the previous frame's (the module where the function is
    called) locals mapping and update it.

    The modified methods and functions are listed in the `sarus/config.yaml`
    file.
    """
    previous_frame = inspect.currentframe().f_back
    local_vars = previous_frame.f_locals
    module_name = local_vars["__name__"]
    module_conf = module_config(module_name)
    if module_conf is None:
        return
    # for documentation
    if use_parent_module:
        doc_module_name = get_parent_module_name(module_name)
    else:
        doc_module_name = module_name

    # Registering module functions
    global _registered_functions
    functions = module_conf.get("sarus_functions", {})
    for fn_name, fn_code_name in functions.items():
        local_vars[fn_name] = create_function(
            fn_code_name, module_name, fn_name, doc_module_name
        )

    # Registering explicit evaluation functions
    explicit_eval_fns = module_conf.get("explicit_eval", [])
    for fn_name in explicit_eval_fns:
        fn_obj = local_vars[fn_name]
        local_vars[fn_name] = explicit_sarus_eval(fn_obj)

    # Registering classes methods
    global _registered_methods
    classes = module_conf.get("classes", {})
    for class_name, methods in classes.items():
        class_obj = local_vars[class_name]
        for mth_name, mth_code_name in methods.items():
            setattr(
                class_obj,
                mth_name,
                create_method(
                    mth_code_name,
                    module_name,
                    class_name,
                    mth_name,
                    doc_module_name,
                ),
            )


def serialize_external(
    code_name: str,
    *args: Union[DataSpecWrapper, Any],
    **kwargs: Union[DataSpecWrapper, Any],
) -> st.DataSpec:
    """This function registers a new dataspec.

    Some arguments are instances of DataSpecWrapper and others are
    just Python object.
    """
    args = [convert_container(arg) for arg in args]
    kwargs = {
        eval(name): convert_container(arg) for name, arg in kwargs.items()
    }
    sarus_default_output = kwargs.pop(SARUS_DEFAULT_OUTPUT, None)
    sarus_types = (DataSpecWrapper, TracedFunction)
    py_args = {
        i: arg
        for i, arg in enumerate(args)
        if not isinstance(arg, sarus_types)
    }
    ds_args_pos = [
        i for i, arg in enumerate(args) if isinstance(arg, sarus_types)
    ]
    ds_arg_types = {
        i: str(arg.__wraps__)
        if isinstance(arg, DataSpecWrapper)
        else "TracedFunction"
        for i, arg in enumerate(args)
        if isinstance(arg, sarus_types)
    }
    ds_args = [
        arg.dataspec(DataSpecVariant.USER_DEFINED)
        if isinstance(arg, DataSpecWrapper)
        else arg.transform()
        for arg in args
        if isinstance(arg, sarus_types)
    ]
    py_kwargs = {
        name: arg
        for name, arg in kwargs.items()
        if not isinstance(arg, sarus_types)
    }
    ds_kwargs = {
        name: arg.dataspec(DataSpecVariant.USER_DEFINED)
        if isinstance(arg, DataSpecWrapper)
        else arg.transform()
        for name, arg in kwargs.items()
        if isinstance(arg, sarus_types)
    }
    ds_kwargs_types = {
        name: str(arg.__wraps__)
        if isinstance(arg, DataSpecWrapper)
        else "TracedFunction"
        for name, arg in kwargs.items()
        if isinstance(arg, sarus_types)
    }
    transform = external(
        id=code_name,
        py_args=py_args,
        py_kwargs=py_kwargs,
        ds_args_pos=ds_args_pos,
        ds_types={**ds_arg_types, **ds_kwargs_types},
        sarus_default_output=sarus_default_output,
    )
    new_dataspec = transform(*ds_args, **ds_kwargs)

    # Compute the traced transform if tracing
    tr_args = [
        arg.traced_transform()
        for arg in args
        if isinstance(arg, DataSpecWrapper)
    ]
    tr_kwargs = {
        name: arg.traced_transform()
        for name, arg in kwargs.items()
        if isinstance(arg, DataSpecWrapper)
    }
    is_not_traced = [tr is None for tr in tr_args] + [
        tr is None for tr in tr_kwargs.values()
    ]
    if all(is_not_traced):
        traced_transform = None
    elif any(is_not_traced):
        raise ValueError("Some inputs are traced and other are not.")
    else:
        traced_transform = transform(*tr_args, **tr_kwargs)

    return new_dataspec, traced_transform


def _sarus_op(
    code_name: str,
    inplace: bool = False,
    register: bool = False,
    is_property: bool = False,
    use_parent_module: bool = False,
) -> Callable:
    """Parametrized decorator to register a Sarus external op."""

    def parametrized_wrapper(ops_fn: Callable) -> Callable:
        @wraps(ops_fn)
        def wrapper_fn(
            *args: Union[DataSpecWrapper, Any],
            **kwargs: Union[DataSpecWrapper, Any],
        ) -> DataSpecWrapper:
            new_dataspec, traced_transform = serialize_external(
                code_name, *args, **kwargs
            )
            context: LocalSDKContext = global_context()

            new_dataspec_wrapper = context.wrapper_factory().create(
                new_dataspec
            )
            if isinstance(new_dataspec_wrapper, DataSpecWrapper):
                new_dataspec_wrapper._set_traced_transform(traced_transform)

            if inplace:
                self: DataSpecWrapper = args[0]  # TODO check semantic
                self._set_dataspec(new_dataspec)
                if isinstance(self, DataSpecWrapper):
                    self._set_traced_transform(traced_transform)

            return new_dataspec_wrapper

        if is_property:
            wrapper_fn = property(wrapper_fn)

        if register:
            wrapper_fn = register_method(
                wrapper_fn, code_name, use_parent_module=use_parent_module
            )

        return wrapper_fn

    return parametrized_wrapper


sarus_method = partial(_sarus_op, register=True, is_property=False)
sarus_property = partial(_sarus_op, register=True, is_property=True)


def sarus_custom_method(code_name: str) -> Callable:
    """Decorator to simply register the method in the documentation.

    The the implementation details are left to the SDK.
    """

    def parametrized_wrapper(method: Callable) -> Callable:
        return register_method(method, code_name)

    return parametrized_wrapper


def sarus_init(code_name: str, use_parent_module: str = False) -> Callable:
    """Decorator to initialize DataSpecWrapper classes from ops."""

    def parametrized_wrapper(ops_fn: Callable) -> Callable:
        @wraps(ops_fn)
        def init_fn(
            self: DataSpecWrapper,
            *args: Union[DataSpecWrapper, Any],
            **kwargs: Union[DataSpecWrapper, Any],
        ) -> None:
            new_dataspec, traced_transform = serialize_external(
                code_name, *args, **kwargs
            )
            self._set_dataspec(new_dataspec)
            self._set_traced_transform(traced_transform)

        init_fn = register_method(
            init_fn, code_name, use_parent_module=use_parent_module
        )

        return init_fn

    return parametrized_wrapper


def create_function(
    code_name: str,
    module_name: str,
    fn_name: str,
    doc_module_name: str = None,
    inplace: bool = False,
) -> Callable:
    """Create an op and register it as a function in a module."""
    global _registered_functions
    if doc_module_name is None:
        doc_module_name = module_name
    _registered_functions.append((doc_module_name, fn_name, code_name))

    @_sarus_op(code_name=code_name, inplace=inplace)
    def dummy_fn(*args, **kwargs): ...

    return dummy_fn


def create_method(
    code_name: str,
    module_name: str,
    class_name: str,
    method_name: str,
    doc_module_name: str = None,
    inplace: bool = False,
) -> Callable:
    """Create an op and register it as a method of a class."""
    global _registered_methods
    if doc_module_name is None:
        doc_module_name = module_name
    _registered_methods.append(
        (doc_module_name, class_name, method_name, code_name)
    )

    @_sarus_op(code_name=code_name, inplace=inplace)
    def dummy_fn(*args, **kwargs): ...

    return dummy_fn


def create_lambda_op(code_name: str, inplace: bool = False) -> Callable:
    """Create an op and do not register it."""

    @_sarus_op(code_name=code_name, inplace=inplace)
    def dummy_fn(*args, **kwargs): ...

    return dummy_fn


@_sarus_op(code_name="std.LEN")
def length(__o: object): ...


@_sarus_op(code_name="std.INT")
def integer(__o: object): ...


@_sarus_op(code_name="std.FLOAT")
def floating(__o: object): ...


def explicit_sarus_eval(func):
    """Decorator to explicitly collect Dataspec's values before calling."""

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        args = [eval(arg) for arg in args]
        kwargs = {key: eval(val) for key, val in kwargs.items()}
        return func(*args, **kwargs)

    return wrapped_func


def init_wrapped(wrapper_class):
    """Define the constructor to return a wrapped instance."""
    assert issubclass(wrapper_class, DataSpecWrapper)

    def __new__(cls, *args, **kwargs):
        return wrapper_class.__wraps__(*args, **kwargs)

    wrapper_class.__new__ = staticmethod(__new__)

    return wrapper_class



# flake8: noqa
from importlib import import_module

import numpy as np
from numpy import *
from numpy import abs

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import init_wrapped, register_ops

from .scalars import *

random = import_module("sarus.numpy.random")


@init_wrapped
class ndarray(DataSpecWrapper[np.ndarray]): ...


register_ops()



from pathlib import Path

import pandas as pd
from sarus_data_spec.manager.ops.processor.external import (
    DP_TRANSFORMS,
    PUP_TRANSFORMS,
)

from sarus.utils import _registered_functions  # noqa: F401
from sarus.utils import _registered_methods  # noqa: F401


def op_list() -> pd.DataFrame:
    """Return the list of ops in a pandas DataFrame."""
    global _registered_functions, _registered_methods

    methods = pd.DataFrame.from_records(_registered_methods)
    methods.columns = ["module", "class", "method", "code"]

    functions = pd.DataFrame.from_records(_registered_functions)
    functions.columns = ["module", "function", "code"]

    all_items = methods.append(functions, ignore_index=True)

    all_items["pup"] = all_items.code.isin(PUP_TRANSFORMS)
    all_items["dp"] = all_items.code.isin(DP_TRANSFORMS)
    all_items = all_items.sort_values(
        ["module", "class", "method", "function", "code"]
    )
    return all_items


def to_markdown(op_list: pd.DataFrame) -> str:
    """Generate the list of registered operations.

    NB: This does not list DataspecWrappers without any operations declared.
    """
    lines = []
    for mod_name, mod_df in op_list.groupby(by="module"):
        lines.append(f"\n# {mod_name}")

        fns = (
            mod_df.loc[:, ["function", "code"]]
            .dropna()
            .sort_values(by="function")
        )

        if len(fns) > 0:
            lines.append("\n## Functions")
            lines += list(map(lambda x: f"- `{x}`", fns.function))

        for class_name, class_df in mod_df.groupby("class"):
            if class_name == "DataSpecWrapper":
                class_name = "Generic Operations"
            lines.append(f"\n## {class_name}")
            methods = (
                class_df.loc[:, ["method", "code"]]
                .dropna()
                .sort_values(by="method")
            )
            lines += list(map(lambda x: f"- `{x}`", methods.method))

    return "\n".join(lines)


def to_markdown_table(op_list: pd.DataFrame) -> str:
    """Generate the op list as a Markdown table."""
    functions = op_list[op_list["class"].isna()]
    methods = op_list[~op_list["class"].isna()]
    methods_full_name = methods[["module", "class", "method"]].apply(
        lambda x: ".".join(x), axis=1
    )
    functions_full_name = functions[["module", "function"]].apply(
        lambda x: ".".join(x), axis=1
    )
    op_list["Supported ops"] = pd.concat(
        [methods_full_name, functions_full_name]
    ).apply(lambda x: f"`{x}`")
    table_data = op_list[["Supported ops", "pup", "dp", "code"]]
    table_data["PUP condition"] = table_data.apply(
        lambda row: PUP_TRANSFORMS.get(row.code, ""), axis=1
    )
    table_data["DP condition"] = table_data.apply(
        lambda row: DP_TRANSFORMS.get(row.code, ""), axis=1
    )
    table_data = table_data.drop("code", axis=1)
    table_data.columns = [
        "Supported ops",
        "PUP",
        "DP",
        "PUP condition",
        "DP condition",
    ]
    table_data = table_data[
        ["Supported ops", "PUP", "PUP condition", "DP", "DP condition"]
    ]
    table_data = table_data.replace({True: "Yes", False: "No"})
    table_data = table_data.sort_values("Supported ops")
    return table_data.to_markdown(index=False)


def to_interactive_html(op_list: pd.DataFrame) -> str:
    """The Jinja template was created using an HTML export of an itables DataFrame."""
    from jinja2 import Environment, FileSystemLoader

    environment = Environment(
        loader=FileSystemLoader(Path(__file__).parent), autoescape=True
    )
    template = environment.get_template("template.html")
    return template.render(data=op_list.astype(str).values.tolist())


if __name__ == "__main__":
    import argparse

    default_output_dir = (
        Path(__file__).parent.parent.parent / "docs" / "sdk_documentation"
    )

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        "-o",
        dest="output_dir",
        default=str(default_output_dir),
    )
    args = parser.parse_args()
    output_dir = Path(args.output_dir)

    all_ops = op_list()

    # All ops as markdown
    with open(output_dir / "op_list.md", "w") as f:
        f.write(to_markdown(all_ops))

    # All ops as markdown table
    with open(output_dir / "op_table.md", "w") as f:
        f.write(to_markdown_table(all_ops))

    # All ops as interactive HTML
    with open(output_dir / "interactive_op_list.html", "w") as f:
        f.write(to_interactive_html(all_ops))



"""Sarus Tensorflow package documentation."""

from sarus.utils import register_ops

try:
    from .xgboost import XGBClassifier
except NameError:
    pass
else:
    __all__ = ["XGBClassifier"]

    register_ops()



# flake8: noqa
from sarus.utils import register_ops

try:
    from shap.plots import *
except ModuleNotFoundError:
    pass

from shap.plots import (
    bar,
    beeswarm,
    benchmark,
    decision,
    embedding,
    force,
    group_difference,
    heatmap,
    image,
    image_to_text,
    initjs,
    monitoring,
    partial_dependence,
    scatter,
    text,
    violin,
    waterfall,
)

register_ops()



from __future__ import annotations

from functools import partial
from typing import Any, Optional

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import xgboost
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing

sarus_method = partial(sarus_method, use_parent_module=True)
sarus_init = partial(sarus_init, use_parent_module=True)


class XGBClassifier(DataSpecWrapper[xgboost.XGBClassifier]):
    @sarus_init("xgboost.XGB_CLASSIFIER")
    def __init__(
        self,
        *,
        objective="binary:logistic",
        use_label_encoder: Optional[bool] = None,
        _dataspec=None,
        **kwargs: Any,
    ): ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y, sample_weight=None): ...


class XGBRegressor(DataSpecWrapper[xgboost.XGBRegressor]):
    @sarus_init("xgboost.XGB_REGRESSOR")
    def __init__(
        self, *, objective="reg:squarederror", _dataspec=None, **kwargs
    ): ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y, sample_weight=None): ...


register_ops(use_parent_module=True)



"""Copyright 2023 Sarus SAS.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Sarus library to leverage sensitive data without revealing them.

This lib contains classes and method to federate different Sarus objects
to process them simultaneously.
"""

from __future__ import annotations

import logging
import types
import typing as t

import sarus
from sarus import Dataset
from sarus.context.typing import LocalSDKContext
from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.pandas.dataframe import DataFrame

SILENCED_IPYTHON_METHODS = [
    "_ipython_canary_method_should_not_exist_",
    "_ipython_display_",
    "_repr_mimebundle_",
    "_repr_html_",
    "_repr_markdown_",
    "_repr_svg_",
    "_repr_png_",
    "_repr_pdf_",
    "_repr_jpeg_",
    "_repr_latex_",
    "_repr_json_",
    "_repr_javascript_",
]

T = t.TypeVar("T")
W = t.TypeVar("W", bound=DataSpecWrapper, covariant=True)


CMP_INTERFACE = {
    "__ge__",
    "__gt__",
    "__le__",
    "__lt__",
    "__ne__",
}


def get_interface(obj: t.Any, parent: t.Any = object) -> t.Set[str]:
    x, y = map(set, map(dir, (obj, parent)))
    return (x - y - {"__dict__"}) | (CMP_INTERFACE & y)


def federated_functor(name) -> t.Callable:
    """Functor to automatically vectorize a method.

    It maps a method from the base class based on its name to a vectorized
    function to be used as a method.

    We have the diagram:

    Vec[X]------------------>Vec[Y]
      |           F            ^
      |                        |
      |federate⁻¹     federate |
      |                        |
      v           f            |
      X----------------------->Y

    TODO: vectorizing functions in general; the output is then the following:
      (⨂f)(U) -> ⨂(f(U)) ~ (f1(U),...,fn(U)).
    More generally, we currently lack the tools to define:
      - (⨂f)(⨂U) -> ⨂(f(U)) ~ (f1(U1),...,fn(Un))
      - f⨂: f -> ((⨂U,V) -> ⨂(f(U,V))) ~ (f(U1,V),...,f(Un,V))
      - f⨂: f -> ((⨂U,⨂V) -> ⨂(f(U,V))) ~ (f(U1,V1),...,f(Un,Vn))
    which requires analyzing args and kwargs, and adding implementations
    for Python builtins.

    TODO: it lacks the support for aligning Federated parameters.
    This means that we have:
      federate([1, 2]) < 1.5 == federate([1 < 1.5, 2 < 1.5])
                             == federate([False, True])
    but we do not have:
      federate([1, 2]) < federate([0, 4]) == federate([1 < 0, 2 < 4])
                                          == federate([False, True])
    Instead we have: NotImplemented.
    NB: `==` is not working for that very same reason !
    """

    def f(self, *args, **kwargs):
        return federate(
            [
                m(*args, **kwargs)
                for m in (getattr(item, name) for item in self.__items__)
            ]
        )

    f.__name__ = name
    return f


def clone_interface(
    federating_class: t.Type,
    federated_class: t.Type,
    base_class: t.Type = object,
) -> None:
    """Adds to the Federating class the methods of the federated.

    This modifies the Federating class in-place.
    Some methods such as `__lt__` and such have to be defined at the
    class-level for some synthetic mechanisms to work, so we cannot rely
    exclusively on `getattr`.

    TODO: a better cloning of properties/descriptors
    see: https://docs.python.org/3/library/inspect.html#inspect.getattr_static
    Currently, properties depend on the getattr implementation
    They could be defined with the dct of the MetaClass and `property(method)`
    but for extensive coverage the issue of the setter and deleter are to
    be investigated as well.
    """
    federating_class.__wrapper_interface__ = get_interface(
        federated_class, base_class
    )
    for name in federating_class.__wrapper_interface__:
        try:
            m = getattr(federated_class, name)
            is_callable = callable(m)
        except Exception:
            is_callable = False

        method = federated_functor(name)
        if is_callable:
            setattr(federating_class, name, method)


class FederatedMeta(type):
    """Metaclass to federate any kind of objects."""

    __federated_instances__: t.Dict[t.Type, t.Type] = {}

    def __new__(cls, name, bases, dct):
        """Creation of a new Federated class."""
        orig_bases = dct["__orig_bases__"]
        federated_class = t.get_args(orig_bases[0])[0]

        new_federating_class = super().__new__(cls, name, bases, dct)
        new_federating_class.__python_type__ = federated_class
        if isinstance(federated_class, type):
            if federated_class in FederatedMeta.__federated_instances__:
                return FederatedMeta.__federated_instances__[federated_class]
            elif issubclass(federated_class, DataSpecWrapper):
                """
                # clone_interface(
                #     new_federating_class, federated_class, DataSpecWrapper
                # )
                # the default behavior of federated wrappers is a fallback to
                # the first element, which is implemented in its `getattr`
                # only **implemented** methods can be vectorized
                """
                pass
            else:
                clone_interface(new_federating_class, federated_class)
            FederatedMeta.__federated_instances__[federated_class] = (
                new_federating_class
            )
        return new_federating_class


class Vectorized(t.Generic[T], metaclass=FederatedMeta):
    """Base class for federated analytics.

    This class groups calls to a list of objects in a duck-typing way.
    We write ⨂P = (P1,...,Pn) our federated collection with n partitions in the docs.
    """

    def __init__(self, items: t.List[T]) -> None:
        self.__items__: t.List[T] = list(items)

    def __repr__(self) -> str:
        items = "\n".join(f"  - {item}" for item in self.__items__)
        return f"{type(self).__name__}:\n{items}"

    def __str_(self) -> str:
        return repr(self)

    def __getattr__(self, name) -> Vectorized:
        """Default behavior is to return a Federate object with the output.

        The output of this function verifies (⨂P).f == ⨂(P.f) ~ (P1.f,...,Pn.f)
        """
        return federate([getattr(item, name) for item in self.__items__])


class WithContext(t.Generic[T]):
    """Minimal wrap to keep safety of execution.

    Any output of the default behavior of FederatedWrapper is wrapped with it,
    so that call is run through the correct context.

    The output can be a method, a closure with a dataspec, a property, so we
    must cover many situation. The solution here is to always return a
    WithContext object.

    TODO: better knowledge of class cloning to only protect known class
    attributes, methods and properties, using a Metaclass.
    """

    def __init__(
        self,
        wrapper: DataSpecWrapper,
        context: t.Optional[LocalSDKContext] = None,
    ) -> None:
        self.context = (
            context if context is not None else wrapper.dataspec().context()
        )
        self.wrapped = wrapper

    def __call__(self, *args, **kwargs) -> t.Any:
        with self.context:
            res = self.wrapped(*args, **kwargs)
        return WithContext(res, self.context)

    def __getattr__(self, name: str) -> t.Any:
        with self.context:
            res = getattr(self.wrapped, name)
        return WithContext(res, self.context)


class FederatedWrapper(Vectorized[W]):
    """Base class for federated DataSpecWrappers.

    It handles the dataspec specifics to make the federated computations work.
    """

    def __map_wrappers__(
        self, fun: t.Callable[[W], t.Any]
    ) -> t.Iterator[t.Any]:
        """Main entry to operate on the DataspecWrappers.

        This iterator yields the output of the function applied to the datasets
        after ensuring the suitable context has been set.
        """
        for item in self.__items__:
            # this is  very permissive to let duck-typing work
            if isinstance(item, DataSpecWrapper):
                with item.dataspec().context():
                    res = fun(item)
            else:
                res = fun(item)
            yield res
        return

    def __federated_eval__(
        self, fun: t.Callable[[W], t.Any]
    ) -> Vectorized[t.Any]:
        """Helper function to apply the function.

        __federated_eval__ has the signature: (⨂W,f) -> ⨂(f(W)) ~ (f(W1),...,f(Wn))
        """
        return federate(list(self.__map_wrappers__(fun)))

    def __sarus_eval__(
        self,
        target_epsilon: t.Optional[float] = None,
        verbose: t.Optional[int] = None,
    ) -> Vectorized[t.Any]:
        """Return a public value for the dataspec.

        We want sarus.eval(⨂W) -> ⨂(sarus.eval(W))

        TODO: __sarus_eval__() means we are getting a result out of our system
        in a plain Python type (not boxed). So the result should probably be
        typed with `U` if `W < DataSpecWrapper[U]` which means we must perform
        reduces here.
        """
        return self.__federated_eval__(
            lambda w: sarus.eval(w, target_epsilon, verbose)
        )

    def __repr__(self) -> str:
        slugnames = ", ".join(
            self.__map_wrappers__(lambda w: w.dataspec()["slugname"])
        )
        return f"{type(self).__name__}({slugnames})"

    def __getattr__(self, name):
        """Default implementation for any op of a FederatedWrapper.

        The current implementation restricts this to: (⨂W).f -> W1.f
        that is: it returns the FederatedWrapper made of the op applied to the first partition.
        """
        if name not in SILENCED_IPYTHON_METHODS:
            logging.warning(
                f"Applying transformation to the first federated object only as '{name}' is not supported in federated version for now. Please tell your account manager if you need it."
            )
        first = next(self.__map_wrappers__(lambda w: getattr(w, name)))
        first_context = self.__items__[0].dataspec().context()
        return WithContext(first, first_context)


class FederatedDataset(FederatedWrapper[Dataset]):
    """A view of a federated dataset as an Arrow iterator."""

    def __init__(self, list_of_datasets: t.List[Dataset]) -> None:
        super().__init__(list_of_datasets)
        slugnames = repr(self)[:-1].replace("FederatedDataset(", "")
        logging.debug(
            f"The datasets {slugnames} have been added to the federated dataset."
        )

    def as_pandas(self) -> FederatedWrapper[sarus.pandas.DataFrame]:
        return self.__federated_eval__(lambda w: w.as_pandas())

    def __getitem__(self, item: t.List[str]) -> Dataset:
        return self.table(item)

    def table(self, table_name: t.List[str]) -> Dataset:
        return self.__federated_eval__(lambda w: w.table(table_name))


class FederatedDataframe(FederatedWrapper[DataFrame]):
    """A view of a federated dataset as a pandas DataFrame."""

    def __getitem__(self, column: str) -> FederatedDataframe:
        return self.__federated_eval__(lambda w: w.__getitem__(column))

    def count(self, *args, **kwargs) -> int:
        return sum(
            self.__federated_eval__(
                lambda w: w.count(*args, **kwargs)
            ).__items__
        )

    def sum(self, *args, **kwargs):
        return sum(
            self.__federated_eval__(lambda w: w.sum(*args, **kwargs)).__items__
        )

    def mean(self, *args, **kwargs):
        S = self.sum(*args, **kwargs)
        N = self.count(*args, **kwargs)
        return S / N[S.index]

    @property
    def shape(self):
        shapes = self.__federated_eval__(lambda w: w.shape).__items__
        return [sum(_[0] for _ in shapes), *shapes[0][1:]]


def federate(items: t.List[T]) -> Vectorized[T]:
    """Main entry function for federating objects.

    It can generate on-the-fly new classes to federate previously unseen
    objects, with a specific care when T < DataSpecWrapper.
    We built a Monad !
    """
    item_type = type(items[0])
    if item_type in FederatedMeta.__federated_instances__:
        return FederatedMeta.__federated_instances__[item_type](items)
    elif issubclass(item_type, DataSpecWrapper):
        federated_base = FederatedWrapper[item_type]
    else:
        federated_base = Vectorized[item_type]
    class_name = f"Federated{item_type.__name__.title()}"
    custom_class = types.new_class(class_name, (federated_base,), {})
    return custom_class(items)



"""Sarus Sklearn package documentation."""
# flake8: noqa

try:
    from sklearn import *

    import sarus.sklearn.cluster as cluster
    import sarus.sklearn.compose as compose
    import sarus.sklearn.decomposition as decomposition
    import sarus.sklearn.ensemble as ensemble
    import sarus.sklearn.impute as impute
    import sarus.sklearn.inspection as inspection
    import sarus.sklearn.linear_model as linear_model
    import sarus.sklearn.metrics as metrics
    import sarus.sklearn.model_selection as model_selection
    import sarus.sklearn.pipeline as pipeline
    import sarus.sklearn.preprocessing as preprocessing
    import sarus.sklearn.svm as svm
except (ModuleNotFoundError, NameError):
    pass
